//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "ezrtsp.h"
#include "ezplay.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
__fastcall EzRtsp::EzRtsp(bool CreateSuspended)
        : TThread(CreateSuspended)
{
}
//---------------------------------------------------------------------------
__fastcall EzRtsp::~EzRtsp()
{
        Sock->Close();
        delete Sock;
}
//---------------------------------------------------------------------------
bool __fastcall EzRtsp::Init() {
        status=EZ_RTSP_CLEAN;
        running=false;
        cseq=1;
        media_audio_trackid=0;
        media_video_trackid=0;
        udp_port1=9000;
        udp_port2=9001;
        udp_port3=9002;
        udp_port4=9003;
        if(Form1->SocketUDP1->LocalPort==0) Form1->SocketUDP1->LocalPort=udp_port1;
        if(Form1->SocketUDP2->LocalPort==0) Form1->SocketUDP2->LocalPort=udp_port2;
        if(Form1->SocketUDP3->LocalPort==0) Form1->SocketUDP3->LocalPort=udp_port3;
        if(Form1->SocketUDP4->LocalPort==0) Form1->SocketUDP4->LocalPort=udp_port4;
        return true;
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::SetSocket(TClientSocket *ts) {
        Sock=ts;
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::Console(AnsiString s) {
        if(Form1 != NULL) Form1->Console(s);
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::SetURL(AnsiString u) {
        media_url=u;
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::Connect() {
        status=EZ_RTSP_CONNECTED;
        if(!running) {
                running=true;
                Resume();
        }
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::Teardown() {
        if(status!=EZ_RTSP_DOWN) { // Si ce n'est pas dj down
                status=EZ_RTSP_TEARDOWN;
                Console("[RTSP] Teardown");
                if(!running) {
                        running=true;
                        Resume();
                }
        }
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::Pause() {
        status=EZ_RTSP_PAUSE;
        Console("[RTSP] Pause");
        if(!running) {
                running=true;
                Resume();
        }
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::Play() {
        if(status==EZ_RTSP_DOWN) {
                status=EZ_RTSP_CONNECTED;
        } else {
                status=EZ_RTSP_PLAY;
        }
        Console("[RTSP] Play");
        if(!running) {
                running=true;
                Resume();
        }
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::Received(AnsiString s)
{
        rbuffer=s;
        if(status==EZ_RTSP_WAIT_DESCRIBE) status=EZ_RTSP_RECEIVED_DESCRIBE;
        else if(status==EZ_RTSP_SETUP_WAIT_AUDIO) status=EZ_RTSP_SETUP_RECEIVED_AUDIO;
        else if(status==EZ_RTSP_SETUP_WAIT_VIDEO) status=EZ_RTSP_SETUP_RECEIVED_VIDEO;
        else if(status==EZ_RTSP_TEARDOWN_WAIT) status=EZ_RTSP_TEARDOWN_RECEIVED;
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::SetFile(AnsiString s)
{
        media_file=s;
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::SetHost(AnsiString s)
{
        media_host=s;
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::Send(AnsiString s)
{
	// Envoi au socket
        s+="User-Agent: EzPlayer\r\n";
        s+="CSeq: "+AnsiString(cseq++)+"\r\n";
	Console("< "+s);
	Sock->Socket->SendText(s+"\r\n");
}
//---------------------------------------------------------------------------
void __fastcall EzRtsp::Execute()
{
        while(running) {
                if(status==EZ_RTSP_CONNECTED) {
                        status=EZ_RTSP_INIT;
                }
                if(status==EZ_RTSP_INIT) {
                        if(Sock->Active) status=EZ_RTSP_DESCRIBE;
                }
                if(status==EZ_RTSP_DESCRIBE) {
                        buffer= "DESCRIBE /"+media_file+" HTTP/1.1" + "\r\n"
                              + "Host: "+media_host                 + "\r\n"
	        	      + "Accept: application/sdp"           + "\r\n";
        		status=EZ_RTSP_WAIT_DESCRIBE;
        		Send(buffer);
                }
                if(status==EZ_RTSP_RECEIVED_DESCRIBE) {
                        Console("[RTSP] DESCRIBE RECEIVED");
                        if(Parse()) {
                                status=EZ_RTSP_SETUP_AUDIO;
                        }
                }
                if(status==EZ_RTSP_SETUP_AUDIO) {
                        Console("-------- SETUP AUDIO");
                        if(media_audio_trackid>0) {
	        	        buffer= "SETUP rtsp://"+media_host+"/"+content_base+"trackID="+AnsiString(media_audio_trackid)+" RTSP/1.0" + "\r\n"
				      + "Transport: RTP/AVP;unicast;client_port="+AnsiString(udp_port1)+"-"+AnsiString(udp_port2)+ "\r\n";
				status=EZ_RTSP_SETUP_WAIT_AUDIO;
                                Send(buffer);
                                Console("[RTSP] WAITING AUDIO SETUP RESPONSE");
                        } else {
                                status=EZ_RTSP_SETUP_VIDEO;
                        }
                }
                if(status==EZ_RTSP_SETUP_RECEIVED_AUDIO) {
                        Console("[RTSP] AUDIO SETUP RECEIVED");
                        if(Parse()) status=EZ_RTSP_SETUP_VIDEO;
                }
                if(status==EZ_RTSP_SETUP_VIDEO) {
                        Console("[RTSP] SETUP VIDEO");
                        if(media_video_trackid>0) {
                                buffer= "SETUP rtsp://"+media_host+"/"+content_base+"trackID="+AnsiString(media_video_trackid)+" RTSP/1.0" + "\r\n"
                                      + "Session: "+media_session                             + "\r\n"
				      + "Transport: RTP/AVP;unicast;client_port="+AnsiString(udp_port3)+"-"+AnsiString(udp_port4)+ "\r\n";
                                status=EZ_RTSP_SETUP_WAIT_VIDEO;
                                Send(buffer);
                                Console("[RTSP] WAITING VIDEO SETUP RESPONSE");
                        } else {
                                status=EZ_RTSP_SETUP_DONE;
                        }
                }
                if(status==EZ_RTSP_SETUP_RECEIVED_VIDEO) {
                        Console("[RTSP] VIDEO RECEIVED");
                        if(Parse()) status=EZ_RTSP_SETUP_DONE;
                }
                if(status==EZ_RTSP_SETUP_DONE) {
                        Console("[RTSP] SETUP DONE");
                        status=EZ_RTSP_PLAY;
                }
                if(status==EZ_RTSP_PLAY) {
                        buffer= "PLAY rtsp://"+media_host+"/"+content_base+" RTSP/1.0" + "\r\n"
                              + "Session: "+media_session                        + "\r\n";
                        Send(buffer);
                        status=EZ_RTSP_WAIT;
                } else if(status==EZ_RTSP_TEARDOWN) {
                        buffer= "TEARDOWN rtsp://"+media_host+"/"+content_base+" RTSP/1.0" + "\r\n"
                              + "Session: "+media_session                        + "\r\n";
                        status=EZ_RTSP_DOWN;
                        Send(buffer);
                } else if(status==EZ_RTSP_PAUSE) {
                        buffer= "PAUSE rtsp://"+media_host+"/"+content_base+" RTSP/1.0" + "\r\n"
                              + "Session: "+media_session                        + "\r\n";
                        status=EZ_RTSP_TEARDOWN_WAIT;
                        Send(buffer);
                } else if(status==EZ_RTSP_TEARDOWN_RECEIVED) {
                        status=EZ_RTSP_WAIT;
                }
                if(status==EZ_RTSP_WAIT) {
                        Console("[RTSP] EZ_RTSP_WAIT");
                        running=false;
                        Suspend();
                } else if (status==EZ_RTSP_DOWN) {
                        Console("[RTSP] EZ_RTSP_DOWN");
                        Form1->Stop();
                        running=false;
                        Suspend();
                }
        }
}
//---------------------------------------------------------------------------
AnsiString __fastcall EzRtsp::ParseSubValue(AnsiString text,AnsiString k, char limit) {
        int pos;
        pos = text.Pos(k);
        if(pos>0) {
                return text.SubString(pos+k.Length(),text.Pos(limit)+1).Trim();
        } else {
                return "";
        }
}
//---------------------------------------------------------------------------
AnsiString __fastcall EzRtsp::ParseValue(AnsiString k, char limit) {
        // Parse le buffer pour trouver la valeur k indique (: ou =)
        int pos;
        pos = rbuffer.Pos(k);
        if(pos>0) {
                rbuffer = rbuffer.SubString(pos+k.Length(),4096); // Consommation
                pos = rbuffer.Pos(limit);
                return rbuffer.SubString(0,pos).Trim();
        } else {
                return "";
        }
}
//---------------------------------------------------------------------------
bool __fastcall EzRtsp::Parse() {
        // Parser la rponse du describe
        if(status==EZ_RTSP_RECEIVED_DESCRIBE) {
                if(rbuffer.SubString(0,12)!="RTSP/1.0 200") {
                        status=EZ_RTSP_DOWN;
                        return false;
                }
                content_base = ParseValue("Content-Base:",'\n');
                media_range = ParseValue("a=range:",'\n');
                for(i=0;i<2;i++) {
                        media_stream = ParseValue("m=",'\n');
                        if(media_stream.SubString(0,5)=="audio") {  // Audio dtecte
                                media_audio_trackid = ParseValue("a=control:trackID=",'\n').ToInt();
                        } else if(media_stream.SubString(0,5)=="video") { // Vido dtecte
                                media_video_trackid = ParseValue("a=control:trackID=",'\n').ToInt();
                        }
                }
                rbuffer="";
                return true;
        } else if(status==EZ_RTSP_SETUP_RECEIVED_AUDIO) {
                media_session = ParseValue("Session:",'\n');
                media_audio_transport = ParseValue("Transport:",'\n');
                media_audio_source = ParseSubValue(media_audio_transport,"source=",';');
                media_audio_server_port = ParseSubValue(media_audio_transport,"server_port=",';');
                Console("! AUDIO SOURCE : "+media_audio_source);
                return true;
        } else if(status==EZ_RTSP_SETUP_RECEIVED_VIDEO) {
                media_video_transport = ParseValue("Transport:",'\n');
                media_video_source = ParseSubValue(media_video_transport,"source=",';');
                media_video_server_port = ParseSubValue(media_video_transport,"server_port=",';');
                Console("! VIDEO SOURCE : "+media_video_source);
                return true;
        } else {
                rbuffer="";
                return false;
        }
}
//---------------------------------------------------------------------------

