//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "ezthread.h"
#include "ezplay.h"
#pragma package(smart_init)

//---------------------------------------------------------------------------

// Ratio antibug
#ifndef VMR_ARMODE_LETTER_BOX
#define VMR_ARMODE_LETTER_BOX 1
#endif

#ifndef VMR_ARMODE_NONE
#define VMR_ARMODE_NONE 0
#endif

#ifndef VOLUME_SILENCE
#define VOLUME_SILENCE -10000
#endif

#define EZ_TIMEOUT 250 // Timeout de lecture

#define SEC_TO_MEDIAT(x) ((LONGLONG)(x*10000000)) // Second -> Media Time

//---------------------------------------------------------------------------

__fastcall EzThread::EzThread(bool CreateSuspended)
        : TThread(CreateSuspended)
{
        status=EZ_MEDIA_STOPPED;
}
//---------------------------------------------------------------------------
void __fastcall EzThread::Console(AnsiString s) {
        if(Form1 != NULL) Form1->Console(s);
}
//---------------------------------------------------------------------------

// Fonction initialisant le mode windowless = pas de nouvelle fentre
// ouverte pour la vido mais lecture dans un composant courant
HRESULT InitWindowlessVMR (
    HWND hwndApp,                  // Handle de la fentre/du composant pour l'overlay
    IGraphBuilder* pGraph,         // Pointeur vers le graph manager
    IVMRWindowlessControl9** ppWc  // Reoit le pointeur vers la VMR
    )
{
    if (!pGraph || !ppWc) return E_POINTER;
    IBaseFilter* pVmr = NULL; 
    IVMRWindowlessControl9* pWc = NULL;
    // On cre la VMR
    HRESULT hr = CoCreateInstance(CLSID_VideoMixingRenderer9, NULL,
        CLSCTX_INPROC, IID_IBaseFilter, (void**)&pVmr); 
    if (FAILED(hr)) {
        return hr;
    }

    // On ajoute la VMR au graph
    hr = pGraph->AddFilter(pVmr, L"Video Mixing Renderer");
    if (FAILED(hr)) {
        pVmr->Release();
        return hr;
    }
    // Mode de rendu
    IVMRFilterConfig9* pConfig;
    hr = pVmr->QueryInterface(IID_IVMRFilterConfig9, (void**)&pConfig);
    if (SUCCEEDED(hr)) {
        hr = pConfig->SetRenderingMode(0x2); // VMRMode_Windowless
        pConfig->Release();
    }
    if (SUCCEEDED(hr)) {
        // Fentre
        hr = pVmr->QueryInterface(IID_IVMRWindowlessControl9, (void**)&pWc);
        if( SUCCEEDED(hr)) {
            hr = pWc->SetVideoClippingWindow(hwndApp);
            if (SUCCEEDED(hr)) {
                *ppWc = pWc; // Retour du pointeur
            }
            else {
                // Une erreur s'est produite alors on libre
                pWc->Release();
            }
        } 
    }
    pVmr->Release(); 
    return hr; 
}

//---------------------------------------------------------------------------

bool __fastcall EzThread::SetAudio(long vol, long bal)
{
        // Balance
        hr = pAudio->put_Balance(bal);
        if(FAILED(hr)) {
                Console("Balance [ERR]");
        }

        // Volume
        if(mute) { // Si muted, modif du volume "cach"
                lastvolume -= vol;
        } else { // Sinon application directe
                hr = pAudio->put_Volume(-vol);
        }
        if(FAILED(hr)) {
                Console("Volume [ERR]");
                return false;
        } else {
                return true;
        }
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::Init()
{
            pGraph = NULL;
            pControl = NULL;
            pAudio = NULL;
            pSeek = NULL;
            pEvent = NULL;
            pWc = NULL;

            /*IBaseFilter *pRtpSink = NULL;*/
            /* hr = CoCreateInstance(CLSID_RtpSink, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (LPVOID *)&pRtpSink); */

            hwnd = Form1->PanelDisplay->Handle;

            media_duree=0; // Dure
            media_position=0; // Position dans le media

            position_lecture=0; // Position
            position_stop=0; // Position
            position_seek=0; // Pour seeker
            media_rate=1.0; // Vitesse lecture

            decalx=0; // Dcalage X
            decaly=0; // Dcalage Y
            ratio=1;  // VMR_ASPECT_RATIO_MODE

            mute = false;

            // Librairie COM
            hr = CoInitialize(NULL);
            if (FAILED(hr)) {
                Console("Erreur : ne peut pas initialiser la lib COM");
                return false;
            }

            // Cration du filtergraphmanager et interrogation des interfaces
            hr = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER,
                                IID_IGraphBuilder, (void **)&pGraph);
            if (FAILED(hr)) {
                Console("Erreur : Instance Graph non initialise");
                return false;
            }
            hr = pGraph->QueryInterface(IID_IMediaControl, (void **)&pControl);
            if(FAILED(hr)) {
                Console("Erreur : MediaControl");
                return false;
            }
            hr = pGraph->QueryInterface(IID_IMediaSeeking, (void **)&pSeek);
            if(FAILED(hr)) {
                Console("Erreur : MediaSeeking");
                return false;
            }
            hr = pGraph->QueryInterface(IID_IMediaEvent, (void **)&pEvent);
            if(FAILED(hr)) {
                Console("Erreur : MediaEvent");
                return false;
            }
            hr = pGraph->QueryInterface(IID_IBasicAudio, (void **)&pAudio);
            if(FAILED(hr)) {
                Console("Erreur : BasicAudio");
                return false;
            }
            //VideoSourceFilter = IBaseFilter->GetFilterByName("RtpSource.1");

            pSeek->SetTimeFormat(&TIME_FORMAT_MEDIA_TIME); // Unit MediaTime : 100ns

            // Initialisation du mode windowless
            hr = InitWindowlessVMR(hwnd, pGraph, &pWc);
            if (SUCCEEDED(hr)) {
                status=EZ_MEDIA_INIT;
                return true;
            } else {
                return false;
            }
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::ReInit() {
        if(pWc!=NULL) {
                pWc->DisplayModeChanged();
                return true;
        } else {
                return false;
        }
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::UpdateRate(float f) {
        if(pSeek!=NULL) {
                media_rate+=f;
                hr = pSeek->SetRate(media_rate);
                return SUCCEEDED(hr);
        } else {
                return false;
        }
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::Pause()
{
        // PAuse la lecture
        if(status==EZ_MEDIA_PLAYING) {
                hr = pControl->Pause();
                if(SUCCEEDED(hr)) {
                        Suspend();
                        status=EZ_MEDIA_PAUSED;
                        return true;
                } else {
                        return false;
                }
        } else {
                return false;
        }
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::Stop()
{
        // Arrte la lecture
        if(status==EZ_MEDIA_PLAYING || status==EZ_MEDIA_PAUSED) {
                hr = pControl->StopWhenReady(); // ou StopWhenReady
                if(SUCCEEDED(hr)) {
                        // On repositionne le flux au dbut
                        media_position = 0;
                        Form1->SetPosition(media_position);
                        pSeek->SetPositions(&media_position, AM_SEEKING_AbsolutePositioning,
                                            NULL, AM_SEEKING_NoPositioning);
                        pControl->Pause(); // Pour afficher la premire image
                        Terminate();
                        status=EZ_MEDIA_STOPPED;
                        return true;
                } else {
                        return false;
                }
        } else {
                return false;
        }
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::Close()
{
        Stop();
        // Libration des ressources
        if(pWc!=NULL) pWc->Release();
        if(pControl!=NULL) pControl->Release();
        if(pEvent!=NULL) pEvent->Release();
        if(pGraph!=NULL) pGraph->Release();
        if(pSeek!=NULL) pSeek->Release();
        if(pAudio!=NULL) pAudio->Release();
        CoUninitialize();
        return true;
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::Seek(LONGLONG position_lecture)
{
        if(position_lecture <= media_duree) {
                hr = pSeek->SetPositions(&position_lecture,AM_SEEKING_AbsolutePositioning,NULL,AM_SEEKING_NoPositioning);
                if(SUCCEEDED(hr)) {
                        return true;
                } else {
                        Console("Erreur : Seek");
                        return false;
                }
        } else {
                Console("Position hors limites");
                return false;
        }
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::Play()
{
        // Lance la lecture
        if(status==EZ_MEDIA_READY) {
                // Lance le graph
                if(pControl!=NULL) {
                        hr = pControl->Run();
                        if(SUCCEEDED(hr)) {
                                status=EZ_MEDIA_PLAYING;
                                Resume();
                        }
                } else {
                        Console("Erreur : Pas de contrle");
                }
        } else if(status==EZ_MEDIA_PAUSED) {
                hr = pControl->Run();
                if(SUCCEEDED(hr)) {
                        status=EZ_MEDIA_PLAYING;
                        Resume();
                }
        } else if(status==EZ_MEDIA_STOPPED) {
                if(pControl!=NULL) {
                        hr = pControl->Run();
                        if(SUCCEEDED(hr)) {
                                status=EZ_MEDIA_PLAYING;
                                Resume();
                        }
                } else {
                        Console("Erreur : Pas de contrle");
                }
        }
        return (status==EZ_MEDIA_PLAYING);
}
//---------------------------------------------------------------------------
void __fastcall EzThread::Execute()
{
        while(status==EZ_MEDIA_PLAYING) {

                // Attend pour la fin du timeout
                // Si INFINITE = on bloque indfiniment
                pEvent->WaitForCompletion(EZ_TIMEOUT, &evCode);

                // Envoi des infos temps  la fentre d'affichage
                pSeek->GetCurrentPosition(&media_position);
                Form1->SetPosition(media_position);

                if(evCode==EC_COMPLETE) {
                    Form1->EndOfMedia();
                } else if(evCode==EC_ERRORABORT) {
                    Console("Erreur de lecture");
                } else if(evCode==EC_USERABORT) {
                    Console("Interruption par l'utilisateur");
                }
        }
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::OpenFile(LPCWSTR lpwstrFile)
{
        if(status==EZ_MEDIA_INIT) {
                // Construction du graph
                hr = pGraph->RenderFile(lpwstrFile, NULL); // Local pour l'instant L"C:\\s.mpg"
                // Render(Pin) pour le stream
                if(FAILED(hr)) {
                        Console("Erreur : RenderFile");
                        return false;
                }

                // Timeline + envoi de la dure  la fentre principale
                pSeek->GetDuration(&media_duree);
                Form1->SetDuration(media_duree);

                // Vitesse lecture
                pSeek->SetRate(media_rate);

                UpdateRect();

                status=EZ_MEDIA_READY;
                return true;
        } else {
                return false;
        }
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::UpdateRect()
{
        // Paramtrage de la taille vido
        hr = pWc->GetNativeVideoSize(&media_width, &media_height, NULL, NULL);
        if(FAILED(hr)) {
                Console("GetNativeVideoSize [ERR]");
                return false;
        }
        SetRect(&rcSrc, 0, 0, media_width, media_height); // Rectangle source
        ::GetClientRect(hwnd, &rcDest); // Aire de destination
        // :: pour niquer borland qui fait chier avec la VCL
        SetRect(&rcDest, 0-decalx, 0-decaly, rcDest.right+decalx, rcDest.bottom+decaly); // Rectangle destination
        pWc->SetVideoPosition(&rcSrc, &rcDest); // Position Vido
        pWc->SetAspectRatioMode(ratio); // Pour conserver le ratio
        return true;
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::RectZoomX(short val) {
        if(val==0) {
                decalx=0;
                ratio=VMR_ARMODE_LETTER_BOX;
        } else {
                decalx+=val;
                ratio=VMR_ARMODE_NONE;
        }
        return UpdateRect();
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::RectZoomY(short val) {
        if(val==0) {
                decaly=0;
                ratio=VMR_ARMODE_LETTER_BOX;
        } else {
                decaly+=val;
                ratio=VMR_ARMODE_NONE;
        }
        return UpdateRect();
}
//---------------------------------------------------------------------------

bool __fastcall EzThread::Mute() {
        if(pAudio!=NULL) {
                hr = pAudio->get_Volume(&volume); // Rcupration volume
                if (hr == E_NOTIMPL) {
                        // Fichier vido only
                        return true; // On ignore et on retourne
                } else if (FAILED(hr)) {
                        Console("Mute [ERR]");
                        return false;
                }
                if(mute) { // Si muted on rtablit le volume mmoris
                        hr = pAudio->put_Volume(lastvolume);
                        mute = false;
                } else { // Sinon on mmorise
                        lastvolume = volume;
                        hr = pAudio->put_Volume(VOLUME_SILENCE);
                        mute = true;
                }
                return SUCCEEDED(hr);
        } else {
                return false;
        }
}
//---------------------------------------------------------------------------

